% Script to do automated measurement of B-H curves using BK Precision'
% Series 8500 electronic load as a constant current source and the Hall
% effect magnetometer. Both have serial ports and we use MATLAB's
% serialport for communication with the devices.
%
% Anton Kruger
% Iowa City, March 2021
%
%  Version 0.9    - First cut at controlling 8500 from MATLAB. Just a few
%                   commands are implemented.
   close all; clear all;
   global g;                  % Global to hold constants for BK's load        
   g.SET_REMOTE_MODE = 1;
   g.SET_REMOTE_MODE = 1;     % Siwtch 8500 to remote/serial control.
   g.SET_LOCAL_MODE = 2;      % Switch 8500 to local (front panel) control.   
   g.SET_LOAD_OFF = 3;        % Turn 8500 load OFF.
   g.SET_LOAD_ON = 4;         % Turn 8500 load ON.
   g.SET_CR_MODE = 5;         % Switch to constant resistance mode.
   g.SET_CR_RESISTANCE = 6;   % Set the R for constant resistance mode.
   g.GET_CR_RESISTANCE = 7;   % Read the constant R.
   g.GET_VALUES = 8;          % Read I, V, P, measured by the 8500.
   g.SET_CV_MODE = 9;         % Switch to constant voltage mode.
   g.SET_CV_VOLTAGE = 10;     % The V for constant voltage mode.
   g.SET_CC_MODE = 11;        % Switch to constant current mode.
   g.SET_CC_CURRENT = 12;     % The I current for constant voltage mode.   

% Open the serial port where the magnetometer is connected to. Send
% a number of 's' characters, which puts the device in a mode where is
% reports the value when it sees any character, rather than automatically
% spitting out numbers.
   nport = serialport("COM1",9600,"Timeout",5);
   good = false;
   for i = 1:5
      fprintf("Switching Hall sense to manual.\n");      
      write(nport,'s','uint8');
      data = nport.readline;
      if (isempty(str2num(data)))
         good = false;
         continue;
      end
      good = true;
      break;
   end
   if (~good) 
      fprintf("Can't connect to Hall sensor\n");
      nport.delete;
      return;
   end

% Open the serial port that the 8500 is connected to.
% 2DO - error checking.
   sport = serialport("COM9",19200,"Timeout",5);

% Open the serial port that the direction switch is connected to.
% 2DO - error checking.
   dport = serialport("COM8",19200,"Timeout",5);
   write(dport,'F','uint8');     % Do not activate the relay.
   for i = 1:5
      write(dport,'T','uint8');
      pause(0.5);
   end
   write(dport,'F','uint8');
     
% Switch 8500 to remote control, and then to constant current mode.
   packet = MakePacket(g.SET_REMOTE_MODE);
   write(sport,packet,'uint8');
   data = read(sport,26,'uint8');  % data(4) is 128 if succesful.
   packet = MakePacket(g.SET_CC_MODE); 
   write(sport,packet,'uint8');
   data = read(sport,26,'uint8');  
   pause(0.5)
      
% Set up the sequence of currents we are going to step through.
   Up = 0:100:3500;
   Down = flip(Up);
   
   Isign = 1;      % Sign of current.
   d = 1;          % Increasing current  
   doSequence(Isign,d,Up,sport,nport);
   
   Isign = 1;      % Sign of current.
   d = 0;          % Decreasing current
   doSequence(Isign,d,Down,sport,nport);
   
   write(dport,'T','uint8');
   pause(0.5);

   Isign = -1;      % Sign of current.
   d = 1;           % Increasing current
   doSequence(Isign,d,Up,sport,nport);
   
   Isign = -1;      % Sign of current.
   d = 0;           % Decreasing current
   doSequence(Isign,d,Down,sport,nport);
   
   write(dport,'T','uint8');
   pause(0.5);
   
   Isign = 1;        % Sign of current.
   d = 1;            % Increasing current  
   doSequence(Isign,d,Up,sport,nport);
   
% Turn the load off.   
   packet = MakePacket(g.SET_LOAD_OFF);
   write(sport,packet,'uint8');
   data = read(sport,26,'uint8');  
   sport.delete;   

% Set Hall sensor to automatic.    
   write(nport,'a','uint8');
   return

function doSequence(Isign,d,Current,sport,nport)
    global g;
    
% The 8500 uses units of 0.1 mA for current. Thus, if we want
% 10 mA, ask for 100, if we want 213 mA,ask for 2130 etc.

   for i = 1:length(Current)
      I = Current(i)*10; 
      packet = MakePacket(g.SET_CC_CURRENT, I);
      write(sport,packet,'uint8');
      pause(0.05)
      data = read(sport,26,'uint8');   
   
      packet = MakePacket(g.SET_LOAD_ON);
      write(sport,packet,'uint8');
      pause(0.05)   
      data = read(sport,26,'uint8'); 
      pause(0.05);
            
% Read Naggas' Hall sensor. Write 's' and the read the response,
% which is the output voltage from the device.
      write(nport,'s','uint8');
      data = nport.readline;      
      fprintf("%d %f %f\n",d,Isign*I/10,str2num(data));
      
% Wait and then repeat.      
      pause(0.5);
   end
   packet = MakePacket(g.SET_LOAD_OFF);
   write(sport,packet,'uint8');
   data = read(sport,26,'uint8');  
end

function [packet] = MakePacket(ptype, value)
   global g;
   packet = uint8(26);
   packet(1:26) = 0;
   packet(1) = 0xAA;
   packet(2) = 0x00;
   
   switch ptype 
      case g.SET_REMOTE_MODE    % Switch to remote (via aserial bus) control.
         packet(3) = 0x20;
         packet(4) = 0x01;

      case g.SET_LOCAL_MODE        % Switch to local (front panel) control.    
         packet(3) = 0x20;
         packet(4) = 0;

      case g.SET_LOAD_OFF          % Turn the DC load off.
         packet(3) = 0x21;
         packet(4) = 0x00;

      case g.SET_LOAD_ON           % Turn the DC load on.
         packet(3) = 0x21;
         packet(4) = 1;

      case g.SET_CR_MODE         % Switch to constant resistance mode.
         packet(3) = 0x28;
         packet(4) = 3;

      case g.SET_CC_MODE
         packet(3) = 0x28;
         packet(4) = 0;

      case g.SET_CC_CURRENT        % Set the load current (CC mode assumed).
          packet(3) = 0x2A;
          v = cast(value,'uint32');
          mask = cast(0xFF,'uint32');      
          packet(4) = cast(bitand(v,mask),'uint8');
          v = bitshift(v,-8);
          packet(5) = cast(bitand(v,mask),'uint8');
          v = bitshift(v,-8);
          packet(6) = cast(bitand(v,mask),'uint8');           
          v = bitshift(v,-8);
          packet(7) = cast(bitand(v,mask),'uint8'); 

      case g.GET_VALUES            % Get the voltage, current, and power.
         packet(3) = 0x5F;
   end
   
   csum = mod(sum(packet),256);
   packet(26) = csum;
end
